package com.achavero.cvf.core;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.achavero.utils.XMLutils;

/**
 * Filter dispatcher del framework
 * Carga todo la configuracin del archivo cvf.xml
 *     (pginas por defecto, interceptores, puntos de peticin de acceso y redirecciones)
 * Al recibir una peticin busca quien la atiende, ejecuta todos los interceptores,
 *     ejecuta la accin, recoje el resultado y redirecciona al jsp necesario
 *     o inicia una nueva peticin
 * @author Adrian Chavero Ramos
 * @version 20 de dic. de 2015
 */
public class CvfFilterDispatcher implements Filter {

	/**
	 * Factory que devuelve la instancia CvfRequestCommand
	 * (objeto que contiene el metodo execute) y es el encargado de ejecutar la accin
	 */
	protected CvfRequestCommandSingletonFactory cvfRequestCommandSingletonFactory;

	/**
	 * Factory que devuelve la instancia del Interceptor
	 * (objeto que contiene el metodo interceptPreProcess que se ejecuta antes de la accin CvfRequestCommand
	 * y el metodo interceptPostProcess que se ejecuta despues de la accin CvfRequestCommand
	 */
	protected CvfInterceptorSingletonFactory cvfInterceptorSingletonFactory;

	/**
	 * Mapa de URIs hacia CvfRequest
	 * (objeto que contiene informacin de como ejectar la peticin,
	 * principalmente la clase que hereda de CvfRequestCommand que ser la accin a ejecutar)
	 */
	protected Map<String, CvfRequest> cvfRequestMap;

	/**
	 * Configuracin de pgina por defecto (error, cvf.xml erroneo)
	 */
	protected Map<String, String> pagesConfig;

	/**
	 * Directorios excluidos del Framework CVF
	 * Todas las peticiones que se hagan sobre estos se tratarn de forma habitual
	 */
	protected List<String> excludedDirectories;

	/**
	 * Mapa de interceptores que se ejecutaran siempre, a no ser que la peticin (CvfRequest)
	 * excluya especificamente este interceptor
	 */
	protected Map<String, CvfInterceptorData> autoInterceptorsMap;

	/**
	 * Mapa de interceptores que se pueden ejecutar para una peticin (CvfRequest)
	 * Cada uno de estos interceptores se tiene que configurar manualmente para configuracin de peticin
	 */
	protected Map<String, CvfInterceptorData> manualInterceptorsMap;

	/**
	 * Lista de extensiones que se excluyen de ser tratadas por el framework CVF
	 * El acceso a estos recursos es directo
	 */
	public static String[] directAccessExtensions = new String[] { ".png", ".jpg", ".gif", ".css", ".js", ".woff",
	        ".woff2", ".ttf", ".svg", ".eot" };

	/**
	 * Constructor del Filter Dispatcher
	 */
	public CvfFilterDispatcher() {
		this.cvfRequestCommandSingletonFactory = new CvfRequestCommandSingletonFactory();
		this.cvfInterceptorSingletonFactory = new CvfInterceptorSingletonFactory();
		this.autoInterceptorsMap = new HashMap<>();
		this.manualInterceptorsMap = new HashMap<>();
		this.cvfRequestMap = new HashMap<>();
		this.excludedDirectories = new ArrayList<>();

		this.pagesConfig = new HashMap<>();
		this.pagesConfig.put("default", "index.jsp");
		this.pagesConfig.put("not_found", "");
		this.pagesConfig.put("cvf_xml_config_error", "");
	}

	@Override
	public void destroy() {
	}

	@Override
	public void doFilter(ServletRequest servletRequest, ServletResponse servletResponse, FilterChain chain)
	        throws IOException, ServletException {

		HttpServletRequest request = (HttpServletRequest) servletRequest;
		HttpServletResponse response = (HttpServletResponse) servletResponse;

		// La parte importante de la URI ex: list.view, addUser.action, style.css, etc..
		String cvfRequestUri = request.getRequestURI().substring(request.getRequestURI().indexOf('/', 1) + 1,
		        request.getRequestURI().length());

		// Para cada directorio excluido, acceder directamente
		for (String directory : this.excludedDirectories) {
			if (cvfRequestUri.startsWith(directory)) {
				chain.doFilter(servletRequest, servletResponse);
				return;
			}
		}

		// URI tiene que terminar con extensin (http://blabla/bla.xxx)
		if (request.getRequestURI().lastIndexOf(".") == -1) {
			CvfFilterDispatcher.goToUri(request, response, this.pagesConfig.get("not_found"));
			return;
		}

		// Si la extensin esta en la lista de extensiones de acceso directo,
		// ir directamente al recurso
		String uriExtension = request.getRequestURI().substring(request.getRequestURI().lastIndexOf("."),
		        request.getRequestURI().length());
		if (Arrays.asList(CvfFilterDispatcher.directAccessExtensions).contains(uriExtension)) {
			chain.doFilter(servletRequest, servletResponse);
			return;
		}

		// Si la URI esta vacia ir a la pgina por defecto
		if (cvfRequestUri.equals("")) {
			CvfFilterDispatcher.goToUri(request, response, this.pagesConfig.get("default"));
			return;
		}

		// Recuperamos el objeto CvfRequest asociado a la URI
		CvfRequest cvfRequest = new CvfRequest();
		if (this.cvfRequestMap.containsKey(cvfRequestUri)) {
			cvfRequest = this.cvfRequestMap.get(cvfRequestUri);
		}
		// Si no hay un CvfRequest asociado a la URI, retornamos error
		if (cvfRequest.getName() == null || cvfRequest.getType() == null) {
			if (this.pagesConfig.get("not_found").equals("")) {
				PrintWriter out = response.getWriter();
				out.print("cvf.xml - URI not found!!");
			} else {
				CvfFilterDispatcher.goToUri(request, response, this.pagesConfig.get("not_found"));
			}
			return;
		}

		// Manejar la peticin con el CvfRequest
		this.handleRequest(cvfRequest, request, response);
	}

	/**
	 * Manejar la peticin con el CvfRequest
	 * @param cvfRequest objeto que encapsula los datos con los que manejar la peticin
	 * @param request
	 * @param response
	 * @throws IOException
	 */
	public void handleRequest(CvfRequest cvfRequest, HttpServletRequest request, HttpServletResponse response)
	        throws IOException {

		// Recuperar el comando/accin asociado a la peticin (CvfRequest)
		CvfRequestCommand command = this.cvfRequestCommandSingletonFactory
		        .getCvfRequestCommand(cvfRequest.getClassName());

		// Crear el objeto invocation con el comando/accin, el fatory de interceptores, el request y el response
		CvfRequestInvocation invocation = new CvfRequestInvocation(cvfRequest, command,
		        this.cvfInterceptorSingletonFactory, request, response);

		String result = invocation.invoke();
		if (result==null) {
			return;
		}

		// Resultado de la invocacin
		String urlResult = cvfRequest.getUrlResult(result);

		// Si el resultado no es correcto, redigerir a la pgina de error
		if (result == "conf_error" || urlResult == null) {
			if (this.pagesConfig.get("cvf_xml_config_error").equals("")) {
				PrintWriter out = response.getWriter();
				out.print("cvf.xml - Error in configuration");
			} else {
				CvfFilterDispatcher.goToUri(request, response, this.pagesConfig.get("cvf_xml_config_error"));
			}
			return;
		}

		// Redigirir a la URI resultado
		CvfFilterDispatcher.goToUri(request, response, urlResult);
	}

	// Redigirir a la URI
	public static void goToUri(HttpServletRequest request, HttpServletResponse response, String uri) {
		String appName = request.getRequestURI().substring(0, request.getRequestURI().indexOf('/', 1));
		String completeUri = appName + (uri.startsWith("/") ? "" : "/") + uri;
		// Si es un jsp, hacer un forward
		if (uri.endsWith(".jsp")) {
			RequestDispatcher rd = request.getRequestDispatcher(uri);
			try {
				rd.forward(request, response);
				return;
			} catch (ServletException | IOException e) {
				e.printStackTrace();
			}
		}
		// Si no es un jsp, hacer una redireccin
		// (se volvera a calcular el CvfRequest,
		// los interceptores se ejecutarn y habr un nuevo resultado)
		try {
			response.sendRedirect(completeUri);
			return;
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * inicializar el Filter Dispatcher a partir del archivo cvf.xml
	 */
	@Override
	public void init(FilterConfig fConfig) throws ServletException {
		String cvfXmlPath = fConfig.getServletContext().getRealPath("/WEB-INF/classes/cvf.xml");

		try {
			File cvfXmlFile = new File(cvfXmlPath);
			DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
			Document document = dBuilder.parse(cvfXmlFile);

			document.getDocumentElement().normalize();

			// Cargar la configuracin de las pginas por defecto
			NodeList pageNodeList = document.getElementsByTagName("page");
			for (int i = 0; i < pageNodeList.getLength(); i++) {
				Node pageNode = pageNodeList.item(i);
				String name = pageNode.getAttributes().getNamedItem("name").getNodeValue();
				String uri = pageNode.getAttributes().getNamedItem("uri").getNodeValue();
				this.pagesConfig.put(name, uri);
			}

			// Cargar los directorios que se excluyen de ser tratados por el framework CVF
			NodeList excludedNodeList = document.getElementsByTagName("excluded");
			for (int i = 0; i < excludedNodeList.getLength(); i++) {
				Node excludedNode = excludedNodeList.item(i);
				String type = excludedNode.getAttributes().getNamedItem("type").getNodeValue();
				String value = excludedNode.getAttributes().getNamedItem("value").getNodeValue();
				if (type.equals("directory")) {
					for (String directory : value.split(" ")) {
						this.excludedDirectories.add(directory);
					}
				}
			}

			// Cargar los interceptores disponibles
			// los que define el usuario + los que define el propio framework
			NodeList interceptorNodeList = document.getElementsByTagName("interceptor");
			for (int i = 0; i < interceptorNodeList.getLength(); i++) {

				Node interceptorNode = interceptorNodeList.item(i);
				XMLutils.removeEmptyNodes(interceptorNode);

				String name = interceptorNode.getAttributes().getNamedItem("name").getNodeValue();
				String className = interceptorNode.getAttributes().getNamedItem("class") != null
				        ? interceptorNode.getAttributes().getNamedItem("class").getNodeValue() : "";
				String intercepting = interceptorNode.getAttributes().getNamedItem("intercepting") != null
				        ? interceptorNode.getAttributes().getNamedItem("intercepting").getNodeValue() : "manual";

				// interceptores que se ejecutaran siempre, a no ser que la peticin (CvfRequest)
				// excluya especificamente este interceptor
				if (intercepting.equals("auto")) {
					this.autoInterceptorsMap.put(name, new CvfInterceptorData(name, className, intercepting));
				}

				// interceptores que se pueden ejecutar para una peticin (CvfRequest)
				// Cada uno de estos interceptores se tiene que configurar manualmente para configuracin de peticin
				if (intercepting.equals("manual")) {
					this.manualInterceptorsMap.put(name, new CvfInterceptorData(name, className, intercepting));
				}

			}

			// Cargar todos los componentes (agrupacin de CvfRequest)
			NodeList componentNodeList = document.getElementsByTagName("component");
			for (int i = 0; i < componentNodeList.getLength(); i++) {

				Node componentNode = componentNodeList.item(i);
				XMLutils.removeEmptyNodes(componentNode);

				// System.out.println(
				// componentNode.getNodeName() + " " +
				// componentNode.getAttributes().getNamedItem("name"));

				// Cargar todos las peticiones CvfRequest
				NodeList requestNodeList = componentNode.getChildNodes();
				for (int j = 0; j < requestNodeList.getLength(); j++) {
					Node requestNode = requestNodeList.item(j);

					String name = requestNode.getAttributes().getNamedItem("name").getNodeValue();
					String type = requestNode.getAttributes().getNamedItem("type").getNodeValue();
					String className = requestNode.getAttributes().getNamedItem("class") != null
					        ? requestNode.getAttributes().getNamedItem("class").getNodeValue() : "";

					// Construir a partir de los datos del xml
					CvfRequest cvfRequest = new CvfRequest(name, type, className);

					// Aadimos todos los interceptores auto
					this.autoInterceptorsMap.keySet().stream().forEach(key -> cvfRequest.addInterceptor(this.autoInterceptorsMap.get(key).className));

					NodeList requestChildNodeList = requestNode.getChildNodes();
					for (int k = 0; k < requestChildNodeList.getLength(); k++) {
						Node childNode = requestChildNodeList.item(k);

						// Configuracin manual de interceptores para cada CvfRequest
						if (childNode.getNodeName().equals("interceptor-ref")) {
							String interceptorRef = (childNode.getAttributes().getNamedItem("name") != null)
							        ? childNode.getAttributes().getNamedItem("name").getNodeValue() : "";

							String exclude = childNode.getAttributes().getNamedItem("exclude") != null
					        		? childNode.getAttributes().getNamedItem("exclude").getNodeValue() : "false";

					        // si no era una exclusin, aadir el interceptor al CvfRequest
			        		if (exclude.equals("false")) {
								if (this.manualInterceptorsMap.containsKey(interceptorRef)
										&& !this.autoInterceptorsMap.containsKey(interceptorRef)) {
									String interceptorClassName = this.manualInterceptorsMap.get(interceptorRef).className;
									cvfRequest.addInterceptor(interceptorClassName);
								}
							// Si es una exlusin, quitar el interceptor que se aadi a partir los auto
							} else {
								String interceptorClassName = this.autoInterceptorsMap.get(interceptorRef).className;
								cvfRequest.removeInterceptor(interceptorClassName);
							}
						}

						// Cargar los resultados posibles del CvfRequest
						if (childNode.getNodeName().equals("result")) {
							String resultName = (childNode.getAttributes().getNamedItem("name") != null)
							        ? childNode.getAttributes().getNamedItem("name").getNodeValue() : "";
							String resultUrl = childNode.getTextContent();
							cvfRequest.putResult(resultName, resultUrl);
						}
					}

					this.cvfRequestMap.put(name + "." + type, cvfRequest);
				}
			}

		} catch (ParserConfigurationException | SAXException | IOException e) {
			System.err.println("init CvfFilterDispatcher: " + e);
		}

	}

}
